/*	$LAAS: endianGen.c,v 1.6 2007/08/02 16:36:58 matthieu Exp $ */

/* 
 * Copyright (c) 2003 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#include "genom-config.h"
__RCSID("$LAAS: endianGen.c,v 1.6 2007/08/02 16:36:58 matthieu Exp $");

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "endianGen.h"
#include "genom.h"
#include "parser.tab.h"

static char *format_type(TYPE_STR *type);
void genEndianVal(FILE *out, DCL_NOM_STR *n, char *muette);
void genEndianEnum(FILE *out, DCL_NOM_LIST *members);

/*** 
 *** Gnration des fonctions d'impression des structures 
 ***/

int
endianGen(FILE *out)
{
    TYPE_LIST *l;
    TYPE_STR *t;
    DCL_NOM_LIST *m, *ltypedefs;
    char buf[80];
    ID_LIST *ln;

    const char *func_header_proto = 
"extern void endianswap_%s ( %s *x, int nDim, int *dims);\n";

    const char *func_header = 
"void endianswap_%s(%s *x, int nDim, int *dims)\n{\n"
"  FOR_EACH_elt(nDim,dims) {\n";


    /* functions */
    script_open(out);
    cat_begin(out);

    fprintf(out,
	    "/*\n"
	    " * Automatically generated by GenoM: do not edit by hand\n"
	    " *\n"
	    " * Copyright (C) 2003-2007 LAAS/CNRS\n"
	    " */\n\n");

    fprintf(out, "#include <portLib.h>\n");
    fprintf(out, "#include \"%sEndian.h\"\n\n", module->name);

    /* Generation des fonctions endian */
    for (l = types; l != NULL; l = l->next) {
	t = l->type;
	if (/* t->used == 0 || */ (t->flags & TYPE_IMPORT)) {
	    continue;
	}

	/* Entete de la fonction */
	fprintf(out, func_header, nom_type1(t), nom_type(t));

	/* Corps de la fonction */
	switch (t->type) {

	  /* Types de base: normalement on ne passe pas la... */
	case CHAR:
	case SHORT:
	case INT:
	case FLOAT:
	case DOUBLE:
	  fprintf(out, "    fprintf(out, \"%s\", *x);\n}\n\n", 
		  format_type(t));
	  break;

	  /* Structure: on traite chaque membre */
	case STRUCT:
	case UNION:
	  for (m = t->members; m != NULL; m = m->next) {
	    sprintf(buf, "((x+elt)->%s)", m->dcl_nom->name);
	    genEndianVal(out, m->dcl_nom, buf);
	  }
	  break;

	  /* Affichage en clair des symboles de l'enum */
	case ENUM:
	  genEndianEnum(out, t->members);
	  break;

	case TYPEDEF:
	  fprintf(stderr, "%s: TYPEDEF dans liste types\n", nomfic);
	  break;
	}

	/* Termine la fonction */
	fprintf(out, 
		"  } END_FOR\n"
		"}\n\n");
    }

    /* 
     * Affichage des typedef: on appelle les fonctions ci-dessus
     * (qui appele directement scan_type ? )
     */
    fprintf(out, "\n/* ======================== ENDIAN DES TYPEDEF ============================= */\n\n");
    for (ltypedefs = typedefs; ltypedefs != NULL; 
	 ltypedefs = ltypedefs->next) {
	if (/* ltypedefs->dcl_nom->type->used == 0
	       ||*/ (ltypedefs->dcl_nom->type->flags & TYPE_IMPORT)) {
	    continue;
	}
	fprintf(out, func_header, ltypedefs->dcl_nom->name, 
		ltypedefs->dcl_nom->name);
	genEndianVal(out, ltypedefs->dcl_nom, "(*(x+elt))");
	fprintf(out, 
		"  } END_FOR\n"
		"}\n\n");
    }

    cat_end(out);
    script_close(out, "server/%sEndian.c", module->name);

    /*
     * Generation header
     */
    script_open(out);
    cat_begin(out);

    fprintf(out,
	    "/*\n"
	    " * Automatically generated by GenoM: do not edit by hand\n"
	    " *\n"
	    " * Copyright (C) 2003 LAAS/CNRS\n"
	    " */\n\n");

    fprintf(out, "#ifndef ENDIANSWAP_%s_H\n", module->NAME);
    fprintf(out, "#define ENDIANSWAP_%s_H\n", module->NAME);


    fprintf(out, "#include <stdio.h>\n");
    fprintf(out, "#include <stdlib.h>\n\n");
    fprintf(out, "\n");
    fprintf(out, "#include <portLib.h>\n");
    fprintf(out, "#include <genom/printScan.h>\n");
    fprintf(out, "#include <genom/h2endian.h>\n");
    fprintf(out, "#include \"%sPosterLib.h\"\n", module->name);

    /* Structures importees d'autres modules */
    for (ln = imports; ln != NULL; ln = ln->next) {
      fprintf(out, "\n#include \"server/%sEndian.h\"\n", ln->name);
    } /* for */

    /* Generation des prototypes des fonctions endian */
    fprintf(out, "\n/* Protoypes */\n\n");
    for (l = types; l != NULL; l = l->next) {
	t = l->type;
	if (/* t->used == 0 || */ (t->flags & TYPE_IMPORT)) {
	    continue;
	}
	fprintf(out, func_header_proto, nom_type1(t), nom_type(t));
    }

    fprintf(out, "\n/* ======================== ENDIAN DES TYPEDEF ============================= */\n\n");
    for (ltypedefs = typedefs; ltypedefs != NULL; 
	 ltypedefs = ltypedefs->next) {
	if (/* ltypedefs->dcl_nom->type->used == 0
	       ||*/ (ltypedefs->dcl_nom->type->flags & TYPE_IMPORT)) {
	    continue;
	}
	fprintf(out, func_header_proto, ltypedefs->dcl_nom->name, 
		ltypedefs->dcl_nom->name);
    }

    fprintf(out, "#endif\n");

    cat_end(out);
    script_close(out, "server/%sEndian.h", module->name);

    return(0);
} /* endianGen */

/*----------------------------------------------------------------------*/

static char *
format_type(TYPE_STR *type)
{
   switch (type->type) {
     case CHAR:
       return("%c");
     case INT:
       return(type->flags & UNSIGNED_TYPE ? "%u" : "%d");
     case FLOAT:
     case DOUBLE:
       return("%f");
     default:
       return(NULL);
    }
} /* format_type */

/*----------------------------------------------------------------------*/

void
genEndianVal(FILE *out, DCL_NOM_STR *n, char *muette)
{
    char *addrstr=NULL;
    int i;
    char *var, *type1;

    /* 
     * affiche le nom de la variable
     */
    dcl_nom_decl(n, &type1, &var);

    /*
     * appel de la fonction d'affichage de cette variable
     */

    /* Type de la donnee a afficher */
    free(type1);
    if(n->flags & STRING)
      type1 = strdup("string");
    else 
      type1 = nom_type1(n->type);

    /* Tableau de valeurs */
    if(n->flags & ARRAY || n->flags & STRING) {
      fprintf(out, 
	      "    { int dims[%d] = {%d",  
	      n->ndimensions, n->dimensions[0]);
      for (i=1;i<n->ndimensions;i++)
	fprintf(out, ",%d", n->dimensions[i]);
      fprintf(out, "};\n  ");
      bufcat(&addrstr, "(%s *)", nom_type(n->type));
    }
    else {
      if (n->pointeur == 0)
	bufcat(&addrstr, "&");
      else
	bufcat(&addrstr, "");
    }

    /* Appel de la fonction endianswap_ */
    if (n->pointeur == 0) {
	fprintf(out, "    endianswap_%s(%s%s, ", 
		type1, addrstr, muette);
    } 
    else {
#ifdef POINTER_ENDIAN_ADDR
	fprintf(out, "    endianswap_addr((void *)(%s), ",
		muette);
#else
	fprintf(out, "    endianswap_%s(", type1);
	for (i = 1; i < n->pointeur; i++) {
	    fprintf(out, "*");
	} /* for */
	fprintf(out, "(%s%s), ", addrstr, muette);
#endif
    }
    if(n->flags & ARRAY || n->flags & STRING)
      fprintf(out, "%d, dims); }\n\n", 
	      n->ndimensions);
    else
      fprintf(out, "0, NULL);\n\n");

    free(var);
    free(type1);
    free(addrstr);
    addrstr = NULL;

} /* genEndianVal */

/*----------------------------------------------------------------------*/

void 
genEndianEnum(FILE *out, DCL_NOM_LIST *members)

{
  fprintf(out, "     endianswap_int((int *)x+elt, 0, NULL);\n");

} /* genEndianEnum */

